import numpy as np
from ..mutual_information import conditional_mutual_information


def _mifs(X, y, lamb=0.0, method='knn', options=None):
    """Select effective features in X on predinting y using
       mutual-information-based feature selection [1]_.
    Parameters
    ----------
    X : array_like, shape (n_samples, d)
        Features.
    y : array_like, shape (n_samples)
        Target. Can be either continuous or discrete.
    lamb: float
        Threshold for independence tests.
    method: str, default 'knn'
        Method used for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    -------
    indices : list
        Indices for the selected features.
    References
    ----------
    .. [1] Brown, Gavin, et al. "Conditional likelihood maximisation:
           a unifying framework for information theoretic feature
           selection." Journal of machine learning research 13.Jan
           (2012): 27-66.
    """
    n, d = X.shape
    selected = []
    while True:
        max_cmi = -np.inf
        not_selected = set(range(d)) - set(selected)
        z = X[:, selected]
        for i in not_selected:
            x = X[:, i]
            cmi = conditional_mutual_information(x, y, z, method, options)
            if max_cmi < cmi:
                max_cmi = cmi
                max_idx = i

        if max_cmi <= lamb or len(selected) == d:
            break

        selected.append(max_idx)

    while True:
        min_cmi = np.inf
        for i in selected:
            x = X[:, i]
            z = X[:, list(set(selected) - set([i]))]
            cmi = conditional_mutual_information(x, y, z, method, options)
            if min_cmi > cmi:
                min_cmi = cmi
                min_idx = i

        if min_cmi >= lamb or len(selected) == 0:
            return selected

        selected.remove(min_idx)



import numpy as np
from ..mutual_information import conditional_mutual_information


def gsmn(X, lamb=0.0, method=None, options=None):
    """Search structure of an MRF using glow-shrink
       Markov networks algorithm [1]_.
    Parameters
    ----------
    X : array, shape (n_samples, d)
        Observations of variables.
    lamb: float
        Threshold for independence tests.
    method: str, default 'knn'
        Method for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    ----------
    adj : array, shape (d, d)
        Estimated structure of an MRF.
    References
    ----------
    .. [1] Bromberg, Facundo, Dimitris Margaritis, and
           Vasant Honavar. "Efficient Markov network
           structure discovery using independence tests."
           Journal of Artificial Intelligence Research 35
           (2009): 449-484.
    """
    n, d = X.shape
    adj = np.zeros([d, d], dtype=bool)
    for i in range(d):
        x = X[:, i]
        non_adj = ~adj[i] & (np.arange(d) != i)
        for j in non_adj.nonzero()[0]:
            y = X[:, j]
            z = X[:, adj[i]]
            cmi = conditional_mutual_information(x, y, z, method=method, options=options)
            if cmi > lamb:
                adj[i, j] = 1
                adj[j, i] = 1

        for j in adj[i].nonzero()[0]:
            other_adj = adj[i] & (np.arange(d) != j)
            y = X[:, j]
            z = X[:, other_adj]
            cmi = conditional_mutual_information(x, y, z, method=method, options=options)
            if cmi <= lamb:
                adj[i, j] = 0
                adj[j, i] = 0

    return adj



import numpy as np
from .dr_ import _mi_dr
from .knn_ import _mi_knn


def mutual_information(X, Y, method='knn', options=None):
    """Estimate mutual information for discrete-continuous mixutres.
    Parameters
    ----------
    X : array_like, shape (n_samples, d_x)
        Variable.
    Y : array_like, shape (n_samples, d_y)
        The other variable.
    method: str, default 'knn'
        Method for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    -------
    mi : float
        Estimated mutual information between each X and Y.
    """
    options = {} if options is None else options
    if X.size == 0 or Y.size == 0:
        return 0
    if np.ndim(X) == 1:
        X = np.reshape(X, [-1, 1])
    if np.ndim(Y) == 1:
        Y = np.reshape(Y, [-1, 1])
    if method == 'dr':
        sigma = options.get('sigma', 1)
        n_bases = options.get('n_bases', 200)
        maxiter = options.get('maxiter', 1000)
        return _mi_dr(X, Y, sigma=sigma, n_bases=n_bases, maxiter=maxiter)
    else:
        k = options.get('k', 3)
        return _mi_knn(X, Y, k)


def conditional_mutual_information(X, Y, Z, method='knn', options=None):
    """Estimate conditional mutual information for discrete-continuous mixutres.
    Parameters
    ----------
    X : array_like, shape (n_samples, d_x)
        Conditioned variable.
    Y : array_like, shape (n_samples, d_y)
        The other conditioned variable.
    Z : array_like, shape (n_samples, d_z)
        Conditioning variable.
    method: str, default 'knn'
        Method for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    -------
    cmi : float
        Estimated conditional mutual information between each X and Y, given Z.
    """
    if Z.size == 0:
        return mutual_information(X, Y, method, options)
    if np.ndim(X) == 1:
        X = np.reshape(X, [-1, 1])
    if np.ndim(Y) == 1:
        Y = np.reshape(Y, [-1, 1])
    if np.ndim(Z) == 1:
        Z = np.reshape(Z, [-1, 1])
    XZ = np.hstack([X, Z])
    mi_xz_y = mutual_information(XZ, Y, method, options)
    mi_y_z = mutual_information(Y, Z, method, options)
    return mi_xz_y - mi_y_z


def mimat(X, method='knn', options=None):
    """Dimension-wise mutual information.
    Parameters
    ----------
    X : array_like, shape (n_samples, d)
        Variable.
    method: str, default 'knn'
        Method for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    -------
    mis : array, shape (d, d)
        Estimated pairwise MIs between dimensions.
    """
    n, d = X.shape
    mis = np.eye(d)
    for i, j in [(i, j) for i in range(d) for j in range(i + 1, d)]:
        x = X[:, [i]]
        y = X[:, [j]]
        mis[i, j] = mutual_information(x, y, method, options)

    mis[mis < 0] = 0
    mis = mis + mis.T
    mis[np.eye(d, dtype=bool)] = np.nan
    return mis


def cmimat(X, method='knn', options=None):
    """Dimension-wise conditional mutual information.
    Parameters
    ----------
    X : array_like, shape (n_samples, d)
        Variable.
    method: str, default 'knn'
        Method for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    -------
    cmis : array, shape (d, d)
        Estimated pairwise CMIs between dimensions, given the other dimensions.
    """
    n, d = X.shape
    cmis = np.eye(d)
    for i, j in [(i, j) for i in range(d) for j in range(i + 1, d)]:
        x = X[:, [i]]
        y = X[:, [j]]
        idx_rest = (np.arange(d) != i) & (np.arange(d) != j)
        z = X[:, idx_rest]
        cmis[i, j] = conditional_mutual_information(x, y, z, method, options)

    cmis[cmis < 0] = 0
    cmis = cmis + cmis.T
    cmis[np.eye(d, dtype=bool)] = np.nan
    return cmis



import numpy as np
from sklearn.covariance import graph_lasso
# In the future, graph_lasso will be renamed to graphical_lasso.


def skeptic(X, alpha):
    """Estimate structure of an MRF with nonparanormal
       SKEPTIC using Spearman’s rho [1]_.
    Parameters
    ----------
    X : array, shape (n_samples, d)
        Observations of variables.
    alpha: float
        Regularization parameter for the graphical lasso.
    Returns
    ----------
    precision : array, shape (d, d)
        Estimated precision (inverse covariance) matrix.
    References
    ----------
    .. [1] Liu, Han, et al. "High-dimensional semiparametric
           Gaussian copula graphical models." The Annals of
           Statistics 40.4 (2012): 2293-2326.
    """
    n, d = X.shape
    indices = np.argsort(X, axis=0)
    rank = np.empty_like(indices)
    for r, idx in zip(rank.T, indices.T):
        r[idx] = np.arange(1, len(X) + 1) - (n + 1) / 2
    rho = rank.T @ rank
    stds = np.sqrt(np.diag(rho))
    rho = rho / stds.reshape(1, -1) / stds.reshape(-1, 1)
    cov = 2 * np.sin(np.pi / 6 * rho)
    cov[np.eye(d, dtype=bool)] = 1
    return graph_lasso(cov, alpha)[1]



import numpy as np
from sklearn.preprocessing import scale
from sklearn.covariance import graph_lasso

from . import skeptic


def _graphical_lasso(X, alpha):
    cov = np.cov(scale(X), rowvar=False)
    return graph_lasso(cov, alpha=alpha)[1]


def _instability(X, estimator, alpha, ratio, rep_num):
    n, p = X.shape
    b = int(ratio * n)
    indices = [np.random.choice(np.arange(n), size=b) for _ in range(rep_num)]
    samples = [scale(sample) for sample in X[indices, :]]
    pres = [estimator(sample, alpha) for sample in samples]
    adjs = np.array([~np.isclose(pre, 0) for pre in pres])
    theta = np.sum(adjs, axis=0) / rep_num
    xi = 2 * theta * (1 - theta)
    d = np.sum(xi) / p / (p - 1)
    return d


def stars(X, estimator, beta=0.1, ratio=None, rep_num=20, lambdas=None, verbose=False):
    n, p = X.shape
    if ratio is None:
        ratio = 10 * (n ** -0.5) if n > 144 else 0.8
    if lambdas is None:
        lambdas = [1e-6, 1e-5, 1e-4, 1e-3, 5e-3, 0.01, 0.03, 0.05, 0.08,
                   0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 1e+6]
    lambdas = sorted(lambdas, reverse=True)  # arrange by descending order

    for i, lamb in enumerate(lambdas):
        d = _instability(X, estimator, lamb, ratio=ratio, rep_num=rep_num)
        if d > beta:
            return lambdas[i - 1]
        if verbose:
            print('[stars] lambda: %f, instability: %f' % (lamb, d))
    return 0


def select(X, method='glasso', beta=0.1, ratio=None, rep_num=20, lambdas=None, verbose=False, return_lambda=False):
    if method == 'glasso':
        estimator = _graphical_lasso
    if method == 'skeptic':
        estimator = skeptic
    lamb = stars(X, estimator, beta=beta, ratio=ratio, rep_num=rep_num, lambdas=lambdas, verbose=verbose)
    if return_lambda:
        return estimator(X, lamb), lamb
    else:
        return estimator(X, lamb)



import numpy as np
from itertools import product

from .sparse_candidate_ import mmpc
from ..mutual_information import mutual_information, conditional_mutual_information


def cyclic(adj):
    path = set()

    def visit(i):
        path.add(i)
        for j in adj[i].nonzero()[0]:
            if j in path or visit(j):
                return True
        path.remove(i)
        return False
    return any(visit(i) for i, _ in enumerate(adj))


def _score_graph(adj, X, method=None, options=None):
    # Should be replaced mutual information test (MIT) criterion,
    # where such criterion has not been developed yet in either
    # purely continuous or discrete-continuous mixed settings.
    n, d = X.shape
    mis = [mutual_information(X[:, i], X[:, adj[i]], method, options) for i in range(d)]
    return np.sum(mis)


def _score_graph_diff(adj, X, idx, operation, method=None, options=None):
    i, j = idx
    if operation == 'add':
        diff = -mutual_information(X[:, i], X[:, adj[i]], method, options)
        adj[i, j] = 1
        diff += mutual_information(X[:, i], X[:, adj[i]], method, options)
        adj[i, j] = 0
    elif operation == 'delete':
        diff = -mutual_information(X[:, i], X[:, adj[i]], method, options)
        adj[i, j] = 0
        diff += mutual_information(X[:, i], X[:, adj[i]], method, options)
        adj[i, j] = 1
    elif operation == 'reverse':
        diff = -mutual_information(X[:, i], X[:, adj[i]], method, options)
        diff -= mutual_information(X[:, j], X[:, adj[j]], method, options)
        adj[i, j] = 0
        adj[j, i] = 1
        diff += mutual_information(X[:, i], X[:, adj[i]], method, options)
        diff += mutual_information(X[:, j], X[:, adj[j]], method, options)
        adj[i, j] = 1
        adj[j, i] = 0
    return diff


def mmhc(X, lamb=0.0, method=None, options=None, verbose=False):
    """Greedily search structure of a Bayesian network using
       max-min hill-climbing algorithm [1]_.
    Parameters
    ----------
    X : array, shape (n_samples, d)
        Observations of variables.
    lamb: float
        Threshold for independence test.
    method: str, default 'knn'
        Method for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    verbose: bool, default False
        Enable verbose output.
    Returns
    ----------
    adj : array, shape (d, d)
        Estimated structure of the Bayesian network.
    References
    ----------
    .. [1] Tsamardinos, Ioannis, Laura E. Brown, and Constantin
           F. Aliferis. "The max-min hill-climbing Bayesian
           network structure learning algorithm." Machine
           learning 65.1 (2006): 31-78.
    """
    n, d = X.shape
    adj = np.zeros([d, d], dtype=bool)
    pc = mmpc(X, lamb, method, options)
    score_prev = -np.inf
    while True:
        score_max = -np.inf
        # Same as "for i in range(d): for j in range(d):"
        for i, j in product(range(d), range(d)):
            if adj[i, j]:
                # Delete
                adj[i, j] = 0
                score = _score_graph(adj, X, method, options)
                if score > score_max:
                    score_max = score
                    idx_max = i, j
                    operation = 'delete'
                # Reverse
                adj[j, i] = 1
                if not cyclic(adj):
                    score = _score_graph(adj, X, method, options)
                    if score > score_max:
                        score_max = score
                        idx_max = i, j
                        operation = 'reverse'
                adj[i, j] = 1
                adj[j, i] = 0
            elif pc[i, j] and not adj[j, i]:
                # Add
                adj[i, j] = 1
                if not cyclic(adj):
                    score = _score_graph(adj, X, method, options)
                    if score > score_max:
                        score_max = score
                        idx_max = i, j
                        operation = 'add'
                adj[i, j] = 0

        if score_max - score_prev <= lamb:
            return adj

        score_prev = score_max
        if verbose:
            print('%s %s: %f' % (operation, idx_max, score_max))
        if operation == 'add':
            adj[idx_max] = 1
        elif operation == 'delete':
            adj[idx_max] = 0
        elif operation == 'reverse':
            adj[idx_max] = 0
            adj[idx_max[::-1]] = 1



import sys
import numpy as np
from scipy.stats import multivariate_normal
from scipy.optimize import minimize
from sklearn.cluster import KMeans


def _normal(X, mean, sigma):
    cov = sigma ** 2 * np.eye(len(mean))
    return multivariate_normal.pdf(X, mean=mean, cov=cov)


def _mi_dr(X, Y, sigma, n_bases, maxiter):
    n, d_x = X.shape
    _, d_y = Y.shape
    b = min(n_bases, n)

    XY = np.hstack([X, Y])
    UV = KMeans(b).fit(XY).cluster_centers_
    U, V = np.split(UV, [d_x], axis=1)
    phi_x = np.array([_normal(X, u, sigma) for u in U])
    phi_y = np.array([_normal(Y, v, sigma) for v in V])
    phi = phi_x * phi_y
    h_x = np.sum(phi_x, axis=1)
    h_y = np.sum(phi_y, axis=1)
    h_xy = np.sum(phi, axis=1)
    h = (h_x * h_y - h_xy) / (n ** 2 - n)

    def fun(alpha):
        return -np.sum(np.log(alpha.dot(phi)))

    def jac(alpha):
        return -phi.dot(1 / alpha.dot(phi))

    bounds = [(0, None)] * b
    constraints = [{'type': 'eq', 'fun': lambda alpha: alpha.dot(h) - 1}]

    alpha0 = np.random.uniform(0, 1, b)
    result = minimize(fun=fun, jac=jac, x0=alpha0, bounds=bounds, constraints=constraints,
                      options={'maxiter': maxiter})

    if not result.success:
        print('Optimization failed: %s' % result.message, file=sys.stderr)

    alpha = result.x
    mi = np.mean(np.log(alpha.dot(phi)))
    return mi



from setuptools import setup, find_packages

setup(name='depynd',
      version='0.3.0',
      description='Evaluating dependencies among random variables.',
      author='Yuya Takashina',
      author_email='takashina2051@gmail.com',
      packages=find_packages(),
      install_requires=[
          'numpy', 'scipy', 'sklearn',
      ],
      url='https://github.com/y-takashina/depynd',
      )
